/*
 * bg_analyzer.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
static bg_visitor_vtbl_t analyzer_vtbl;

int bg_analyzer_create(bg_visitor_t *vis)
{
  /////////////////////////////////////////////////////////////////////////////
  vis->vtbl=&analyzer_vtbl;
  vis->depth=0;

  /////////////////////////////////////////////////////////////////////////////
  return 0;
}

static void bg_analyzer_destroy(bg_visitor_t *vis FFUNUSED)
{
}

#if defined (BG_FILE) // [
///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_dispatch_file(bg_visitor_t *vis FFUNUSED,
    bg_tree_t *tree FFUNUSED)
{
  // nothing to analyze.
  return 0;
}
#endif // ]

#if defined (BG_PRINT) // [
///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_print_prefix(bg_visitor_t *vis, bg_tree_t *tree)
{
  FILE *f=tree->param->result.f;
#if 0 // [
  bg_print_vmt_t *stdout_vtbl=&bg_print_classic_vmt;
#else // ] [
  bg_print_vmt_t *stdout_vtbl=tree->param->print.vtbl;
#endif // ]
  int stdout_prefix=BG_TREE_TYPE_TRACK==tree->vtbl->type||stdout_vtbl->infix;

  if (stdout_prefix) {
    ///////////////////////////////////////////////////////////////////////////
    // the header of a track is always mirrored to the shell/console
    // regardless wether prefix or infix.
    if (stdout_vtbl->head(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing head");
      goto e_print_head_stdout;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_infix;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
e_print_head_file_infix:
e_print_head_stdout:
  return -1;
}

static int bg_analyzer_print_postfix(bg_visitor_t *vis, bg_tree_t *tree)
{
  FILE *f=tree->param->result.f;
#if 0 // [
  bg_print_vmt_t *stdout_vtbl=&bg_print_classic_vmt;
#else // ] [
  bg_print_vmt_t *stdout_vtbl=tree->param->print.vtbl;
#endif // ]
  int stdout_prefix=BG_TREE_TYPE_TRACK==tree->vtbl->type||stdout_vtbl->infix;

  if (!stdout_prefix) {
    if (stdout_vtbl->head(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing head");
      goto e_print_head_stdout;
    }
  }

  if (!tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_postfix;
    }
  }

  if (tree->param->print.vtbl->tail(tree,vis->depth,f)<0) {
    DMESSAGE("printing tail");
    goto e_print_tail_file;
  }

  fflush(f);

  if (stdout!=f) {
    if (tree->param->print.vtbl->tail(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing tail");
      goto e_print_tail_stdout;
    }

    fflush(stdout);
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
e_print_tail_stdout:
e_print_tail_file:
e_print_head_file_postfix:
e_print_head_stdout:
  return -1;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_dispatch_track(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;
#if 0 // [
  FILE *f=tree->param->result.f;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  ++vis->depth; // [

#if defined (BG_PRINT) // [
#if 0 // [
  if (bg_print_classic_vmt.head(tree,vis->depth,stdout)<0) {
    DMESSAGE("printing head");
    goto e_print_head_stdout;
  }
#else // ] [
  if (bg_analyzer_print_prefix(vis,tree)<0) {
    DMESSAGE("printing prefix");
    goto e_print_prefix;
  }
#endif // ]
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (ff_analyzer_loop(&tree->track.analyzer)<0) {
    DMESSAGE("decoding");
    goto e_decode;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_LIB1770) // [
#if defined (BG_PRINT) // [
#if 0 // [
  if (tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
    DMESSAGE("printing tail");
    goto e_print_head_file;
  }

  if (tree->param->print.vtbl->tail(tree,vis->depth,f)<0) {
    DMESSAGE("printing tail");
    goto e_print_tail_file;
  }

  if (stdout!=f&&tree->param->print.vtbl->tail(tree,vis->depth,stdout)<0) {
    DMESSAGE("printing tail");
    goto e_print_tail_stdout;
  }
#else // ] [
  if (bg_analyzer_print_postfix(vis,tree)<0) {
    DMESSAGE("printing postfix");
    goto e_print_postfix;
  }
#endif // ]
#else // ] [
  bg_tree_result(tree,f);
#endif // ]
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  err=0;
//cleanup:
#if 0 // [
#if defined (BG_LIB1770) && defined (BG_PRINT) // [
e_print_tail_stdout:
e_print_tail_file:
e_print_head_file:
#endif // ]
#else // ] [
e_print_postfix:
#endif // ]
e_decode:
#if 0 // [
#if defined (BG_PRINT) // [
e_print_head_stdout:
#endif // ]
#else // ] [
e_print_prefix:
#endif // ]
  --vis->depth; // ]
  return err;
}

static int bg_analyzer_dispatch_album(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;
#if 0 // [
  FILE *f=tree->param->result.f;
#endif // ]
  bg_album_t *album=&tree->album;
  bg_tree_t *cur;

  /////////////////////////////////////////////////////////////////////////////
  ++vis->depth; // [

#if defined (BG_PRINT) // [
  /////////////////////////////////////////////////////////////////////////////
#if 0 // [
  if (tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_infix;
    }
  }
#else // ] [
  if (bg_analyzer_print_prefix(vis,tree)<0) {
    DMESSAGE("printing prefix");
    goto e_print_prefix;
  }
#endif // ]
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  for (cur=album->first;cur;cur=cur->next) {
    if (cur->vtbl->accept(cur,vis)<0) {
      FFVMESSAGE(FFL("analyzing \"%s\""),cur->source.path);
      goto e_child;
    }

#if defined (BG_LIB1770) // [
    if (bg_tree_merge(tree,cur)<0) {
      DMESSAGE("merging");
      goto e_merge;
    }
#endif // ]
  }

#if defined (BG_PRINT) // [
  /////////////////////////////////////////////////////////////////////////////
#if 0 // [
  if (bg_print_classic_vmt.head(tree,vis->depth,stdout)<0) {
    DMESSAGE("printing head");
    goto e_print_head_stdout;
  }

  if (!tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_postfix;
    }
  }

  if (tree->param->print.vtbl->tail(tree,vis->depth,f)<0) {
    DMESSAGE("printing tail");
    goto e_print_tail_file;
  }

  fflush(f);

  if (stdout!=f) {
    if (tree->param->print.vtbl->tail(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing tail");
      goto e_print_tail_stdout;
    }

    fflush(stdout);
  }
#else // ] [
  if (bg_analyzer_print_postfix(vis,tree)<0) {
    DMESSAGE("printing postfix");
    goto e_print_postfix;
  }
#endif // ]
#elif defined (BG_LIB1770) // ] [
  bg_tree_result(tree,f);
  fflush(f);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  err=0;
//cleanup:
#if 0 // [
#if defined (BG_LIB1770) && defined (BG_PRINT) // [
e_print_tail_file:
e_print_tail_stdout:
#endif // ]
#if  defined (BG_PRINT) // [
e_print_head_stdout:
e_print_head_file_postfix:
#endif // ]
#else // ] [
#if defined (BG_PRINT) // [
e_print_postfix:
#endif // ]
#endif // ]
e_merge:
e_child:
#if defined (BG_PRINT) // [
#if 0 // [
e_print_head_file_infix:
#else // ] [
e_print_prefix:
#endif // ]
#endif // ]
  --vis->depth; // ]
  return err;
}

static int bg_analyzer_dispatch_root(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_dispatch_album(vis,tree)<0) {
    DMESSAGE("analyzing");
    goto e_album;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=0;
e_album:
  return err;
}

static bg_visitor_vtbl_t analyzer_vtbl={
#if defined (PBU_DEBUG) // [
  .id="analyzer",
#endif // ]
  .destroy=bg_analyzer_destroy,
#if defined (BG_FILE) // [
  .dispatch_file=bg_analyzer_dispatch_file,
#endif // ]
  .dispatch_track=bg_analyzer_dispatch_track,
  .dispatch_album=bg_analyzer_dispatch_album,
  .dispatch_root=bg_analyzer_dispatch_root,
};
